from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.currency.serializers import CurrencyRepresentationSerializer
from wbcore.contrib.directory.models import BankingContact
from wbportfolio.models import Portfolio, PortfolioPortfolioThroughModel
from wbportfolio.serializers.synchronization import (
    PortfolioSynchronizationRepresentationSerializer,
)


class PortfolioRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:portfolio-detail")
    _detail_preview = wb_serializers.HyperlinkField(reverse_name="wbportfolio:portfolio-detail")

    class Meta:
        model = Portfolio
        fields = ("id", "name", "_detail", "_detail_preview")


class PortfolioModelSerializer(wb_serializers.ModelSerializer):
    _currency = CurrencyRepresentationSerializer(source="currency")
    _depends_on = PortfolioRepresentationSerializer(source="depends_on", many=True)
    _portfolio_synchronization = PortfolioSynchronizationRepresentationSerializer(source="portfolio_synchronization")
    dependent_portfolios = wb_serializers.PrimaryKeyRelatedField(
        many=True, read_only=True, label="Dependency Portfolios"
    )
    _dependent_portfolios = PortfolioSynchronizationRepresentationSerializer(many=True, source="dependent_portfolios")

    @wb_serializers.register_only_instance_resource()
    def cash_management(self, instance, request, user, **kwargs):
        additional_resources = dict()
        if instance.daily_cashflows.exists():
            additional_resources["daily_cashflows"] = reverse(
                "wbportfolio:portfolio-portfoliocashflow-list", args=[instance.id], request=request
            )
        b = BankingContact.objects.filter(wbportfolio_products__id__in=instance.instruments.all().values("id"))
        if b.exists():
            base_url = reverse("wbaccounting:futurecashflow-list", request=request)
            additional_resources[
                "cash_flow"
            ] = f'{base_url}?banking_contact={",".join([str(i) for i in b.distinct("id").values_list("id", flat=True)])}'

        return additional_resources

    @wb_serializers.register_only_instance_resource()
    def additional_resources(self, instance, request, user, **kwargs):
        additional_resources = dict()
        additional_resources["distribution_chart"] = reverse(
            "wbportfolio:portfolio-distributionchart-list", args=[instance.id], request=request
        )
        additional_resources["distribution_table"] = reverse(
            "wbportfolio:portfolio-distributiontable-list", args=[instance.id], request=request
        )
        if instance.assets.exists():
            additional_resources["assets"] = reverse(
                "wbportfolio:portfolio-asset-list", args=[instance.id], request=request
            )
            additional_resources["contributor"] = reverse(
                "wbportfolio:portfolio-contributor-list",
                args=[instance.id],
                request=request,
            )
        if user.profile.is_internal:
            additional_resources["instruments"] = reverse(
                "wbportfolio:portfolio-instrument-list",
                args=[instance.id],
                request=request,
            )

        if instance.portfolio_synchronization:
            additional_resources["resynchronize"] = reverse(
                "wbportfolio:portfolio-resynchronize",
                args=[instance.id],
                request=request,
            )
        additional_resources["dependencyportfolios"] = reverse(
            "wbportfolio:portfolio-dependencyportfolio-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["trade_proposals"] = reverse(
            "wbportfolio:portfolio-tradeproposal-list",
            args=[instance.id],
            request=request,
        )

        additional_resources["preferredclassification"] = reverse(
            "wbportfolio:portfolio-preferredclassification-list",
            args=[instance.id],
            request=request,
        )
        additional_resources["modelcomposition"] = reverse(
            "wbportfolio:portfolio-modelcompositionpandas-list",
            args=[instance.id],
            request=request,
        )

        return additional_resources

    class Meta:
        model = Portfolio
        fields = (
            "id",
            "name",
            "last_synchronization",
            "_depends_on",
            "_portfolio_synchronization",
            "depends_on",
            "portfolio_synchronization",
            "_dependent_portfolios",
            "dependent_portfolios",
            "_currency",
            "currency",
            "_additional_resources",
        )


class ModelPortfolioModelSerializer(PortfolioModelSerializer):
    @wb_serializers.register_only_instance_resource()
    def rebalance(self, instance, request, user, **kwargs):
        return {
            "modelcomposition": reverse(
                "wbportfolio:portfolio-modelcompositionpandas-list",
                args=[instance.id],
                request=request,
            )
        }


class PortfolioPortfolioThroughModelSerializer(wb_serializers.ModelSerializer):
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")
    _dependency_portfolio = PortfolioRepresentationSerializer(source="dependency_portfolio")

    class Meta:
        model = PortfolioPortfolioThroughModel
        fields = (
            "id",
            "_portfolio",
            "portfolio",
            "_dependency_portfolio",
            "dependency_portfolio",
            "type",
        )



import factory
from django_celery_beat.models import CrontabSchedule, PeriodicTask
from wbportfolio.models import (
    PortfolioSynchronization,
    PriceComputation,
    SynchronizationTask,
)


class CrontabScheduleFactory(factory.django.DjangoModelFactory):
    hour = factory.Iterator(range(0, 24))
    minute = factory.Iterator(range(0, 60))

    class Meta:
        model = CrontabSchedule


class PeriodicTaskFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = PeriodicTask

    name = factory.Sequence(lambda n: f"Synchronization Task {n}")
    crontab = factory.SubFactory(CrontabScheduleFactory)


class SynchronizationTaskFactory(PeriodicTaskFactory):
    class Meta:
        model = SynchronizationTask


class PortfolioSynchronizationFactory(SynchronizationTaskFactory):
    propagate_history = False

    class Meta:
        model = PortfolioSynchronization


class PriceComputationFactory(SynchronizationTaskFactory):
    class Meta:
        model = PriceComputation



import logging
from datetime import date, datetime
from decimal import Decimal
from typing import Any, Optional

import numpy as np
import pandas as pd
from celery import shared_task
from celery.canvas import Signature
from django.db import models
from django.db.models import F
from django.utils import timezone
from wbfdm.enums import MarketData
from wbfdm.models.instruments import Instrument
from wbportfolio.models import Portfolio
from wbportfolio.models.transactions.trade_proposals import TradeProposal
from wbportfolio.pms.typing import Portfolio as PortfolioDTO
from wbportfolio.pms.typing import Position as PositionDTO

from .synchronization import SynchronizationTask


def convert_currency(x, val_date, other_currency):
    instrument = Instrument.objects.get(id=x)
    try:
        return instrument.currency.convert(val_date, other_currency)
    except Exception:
        return np.nan


class PortfolioSynchronization(SynchronizationTask):
    is_automatic_validation = models.BooleanField(
        default=True,
        verbose_name="Automatic validation",
        help_text="Set to True if you want to automatically implement proposed positions",
    )
    propagate_history = models.BooleanField(
        default=False,
        verbose_name="Propagate History",
        help_text="If true, when the depends on portfolio changes at a certain date, this method will trigger a synchronization for each date (at the scheduled frequency) from that date to the latest valid date",
    )

    def synchronize(
        self,
        portfolio: models.Model,
        sync_date: date,
        task_execution_datetime: Optional[datetime] = None,
        override_execution_datetime_validity: Optional[bool] = False,
        post_processing: bool = True,
        **kwargs: Any,
    ):
        """
        This function compute the new portfolio composition after synchronization (returns from `_import_method`) for a
        given date and either update or create the portfolio or create a trade proposal given the new portfolio constituent.

        :param portfolio: The portfolio to synchronize the positions from
        :param sync_date: The date at which we need to synchronize the given portfolio
        :param task_execution_datetime: An optional datetime specifying at which time this task was initially executed.
        :param override_execution_datetime_validity: If true, we don't valide `task_execution_datetime`
        :param kwargs: keyword arguments
        """

        initkwargs = {**kwargs, **self.cast_kwargs}
        if not task_execution_datetime:
            task_execution_datetime = timezone.now()
        if portfolio.is_active_at_date(sync_date):
            if self.is_valid_date(task_execution_datetime) or override_execution_datetime_validity:
                if import_res := list(zip(*self._import_method(portfolio, sync_date, **initkwargs))):
                    effective_positions = list(filter(lambda x: x, import_res[0]))
                    target_positions = list(filter(lambda x: x, import_res[1]))
                    if len(target_positions) > 0:
                        target_portfolio = PortfolioDTO(target_positions)

                        effective_portfolio = (
                            PortfolioDTO(effective_positions) if len(effective_positions) > 0 else None
                        )
                        if self.is_automatic_validation:
                            # We process these positions automatically
                            portfolio.import_positions_at_date(
                                target_portfolio, sync_date, post_processing=post_processing
                            )
                        else:
                            trade_proposal, created = TradeProposal.objects.get_or_create(
                                trade_date=sync_date,
                                portfolio=portfolio,
                                defaults={"comment": "Automatic rebalancing"},
                            )
                            trade_proposal.create_or_update_trades(
                                target_portfolio=target_portfolio, effective_portfolio=effective_portfolio
                            )

                        portfolio.last_synchronization = timezone.now()
                        portfolio.save()

            else:
                logging.info(
                    f"Synchronization invalid: {portfolio.name} synchronization with {self.name} was triggered for {sync_date} but date not valid for crontab schedule {str(self.crontab)}"
                )

    def synchronize_as_task_si(self, portfolio: models.Model, sync_date: date, **kwargs: Any) -> Signature:
        """
        Utility function that returns the signature of the synchronize method
        """
        return synchronize_portfolio_as_task.si(self.id, portfolio.id, sync_date, **kwargs)

    def _tasks_signature(self, sync_date: Optional[date] = None, **kwargs: Any) -> Signature:
        """
        Gather all tasks that needs to run under this synchronization job as a list of celery signatures.
        This method is expected to be implemented at each inheriting class.
        :param args: list
        :param kwargs: dict
        :return: list[signature]
        """
        for portfolio in self.portfolios.all():
            portfolio_sync_dates = []
            if sync_date:
                portfolio_sync_dates = [sync_date]
            elif not sync_date and portfolio.assets.exists() and (latest_asset := portfolio.assets.latest("date")):
                portfolio_sync_dates = map(
                    lambda x: x.date(), pd.date_range(latest_asset.date, date.today(), freq="B", inclusive="left")
                )
            for portfolio_sync_date in portfolio_sync_dates:
                if portfolio.is_active_at_date(portfolio_sync_date):
                    yield synchronize_portfolio_as_task.si(self.id, portfolio.id, portfolio_sync_date, **kwargs)

    @classmethod
    def _default_callback(
        cls,
        portfolio: Portfolio,
        sync_date: date,
        portfolio_created: Optional[Portfolio] = None,
        adjusted_weighting: Optional[Decimal] = Decimal(1.0),
        adjusted_currency_fx_rate: Optional[Decimal] = Decimal(1.0),
        is_estimated: Optional[bool] = False,
        portfolio_total_value: Optional[float] = None,
        **kwargs: Any,
    ):
        """Recursively calculates the position for a portfolio

        Arguments:
            portfolio {portfolio.Portfolio} -- The Portfolio on which the assets will be computed
            sync_date {datetime.date} -- The date on which the assets will be computed

        Keyword Arguments:
            portfolio {portfolio.Portfolio} -- The core portfolio from which the computed position are created (default: {None})
            adjusted_weighting {int} -- the adjusted weight of the current level of index (default: {1})
            adjusted_currency_fx_rate {int} -- the adjusted currency exchange rate on the current level of index (default: {1})

        Yields:
            tuple[dict, dict] -- Two dictionaries: One with filter parameters and one with default values
        """
        is_root_position_estimated = False
        if not portfolio_created:
            if portfolio_created := portfolio.primary_portfolio:
                is_root_position_estimated = (
                    portfolio_created.assets.filter(date=sync_date).count() == 1
                    and portfolio_created.assets.filter(date=sync_date, is_estimated=True).count() == 1
                )
        if portfolio_created:
            child_positions = portfolio_created.assets.filter(date=sync_date)
            asset_positions = child_positions.all()
            # Compute the total portfolio value based on the root position child (otherwise the value is passed as
            # parameters in the recursion
            if not portfolio_total_value:
                portfolio_total_value = child_positions.aggregate(tv=models.Sum(F("total_value_fx_portfolio")))["tv"]
                if not portfolio_total_value:
                    portfolio_total_value = portfolio_created.get_total_value(sync_date)
            for position in child_positions:
                if child_portfolio := position.underlying_instrument.portfolio:
                    if child_portfolio.assets.filter(date=sync_date).exists() and position.weighting is not None:
                        asset_positions = asset_positions.exclude(id=position.id)
                        yield from cls._default_callback(
                            portfolio,
                            sync_date,
                            portfolio_created=child_portfolio,
                            adjusted_weighting=position.weighting * adjusted_weighting,
                            portfolio_total_value=portfolio_total_value,
                            adjusted_currency_fx_rate=position.currency_fx_rate * adjusted_currency_fx_rate,
                            is_estimated=False
                            if is_root_position_estimated
                            else (is_estimated and position.is_estimated),
                        )
            df = pd.DataFrame(
                asset_positions.values_list(
                    "currency_fx_rate",
                    "price",
                    "weighting",
                    "shares",
                    "is_estimated",
                    "underlying_instrument",
                    "currency",
                    "exchange",
                ),
                columns=[
                    "currency_fx_rate",
                    "price",
                    "weighting",
                    "shares",
                    "is_estimated",
                    "underlying_instrument",
                    "currency",
                    "exchange",
                ],
            )
            if not df.empty:
                df.currency_fx_rate = df.currency_fx_rate * adjusted_currency_fx_rate
                df.weighting = df.weighting * adjusted_weighting

                df = (
                    df.groupby(["underlying_instrument", "currency", "exchange"], dropna=False)
                    .agg(
                        {
                            "currency_fx_rate": "first",
                            "price": "first",
                            "weighting": "sum",
                            "shares": "sum",
                            "is_estimated": "first",
                        }
                    )
                    .reset_index()
                )
                df[["underlying_instrument", "currency", "exchange"]] = df[
                    ["underlying_instrument", "currency", "exchange"]
                ].astype("object")
                df[["currency_fx_rate", "price", "weighting", "shares"]] = df[
                    ["currency_fx_rate", "price", "weighting", "shares"]
                ].astype("float")

                df["actual_currency_fx_rate"] = df.underlying_instrument.apply(
                    lambda x: convert_currency(x, sync_date, portfolio.currency)
                ).astype("float")
                df["actual_currency_fx_rate"] = df["actual_currency_fx_rate"].fillna(df["currency_fx_rate"])

                df = df.where(pd.notnull(df), None).set_index("underlying_instrument")
                missing_prices = df.loc[df["price"].isnull(), "price"]
                if not missing_prices.empty:
                    prices_df = pd.DataFrame(
                        Instrument.objects.filter(id__in=missing_prices.index).dl.market_data(
                            values=[MarketData.CLOSE], exact_date=sync_date
                        )
                    )
                    if not prices_df.empty:
                        prices_df = prices_df[["close", "instrument_id"]].set_index("instrument_id").astype("float")
                        df.loc[prices_df.index, "price"] = prices_df

                if portfolio_total_value is not None:
                    df["shares"] = (df["weighting"] * float(portfolio_total_value)) / (
                        df["price"] * df["actual_currency_fx_rate"]
                    )
                if is_estimated:
                    df["is_estimated"] = True
                for underlying_instrument, asset_position in df.to_dict("index").items():
                    if (
                        asset_position["weighting"] or asset_position["shares"]
                    ):  # We don't yield empty position (pos with shares and weight equal to 0 or None)
                        # We return the position as a serialized dictionary
                        yield None, PositionDTO(
                            date=sync_date,
                            asset_valuation_date=sync_date,
                            portfolio_created=portfolio_created.id,
                            underlying_instrument=underlying_instrument,
                            instrument_type=Instrument.objects.get(id=underlying_instrument).security_instrument_type,
                            currency=asset_position["currency"],
                            exchange=asset_position["exchange"],
                            shares=asset_position["shares"],
                            price=asset_position["price"],
                            currency_fx_rate=asset_position["actual_currency_fx_rate"],
                            weighting=asset_position["weighting"],
                            is_estimated=asset_position["is_estimated"],
                        )

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbportfolio:portfoliosynchronizationrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


@shared_task(queue="portfolio")
def synchronize_portfolio_as_task(synchronization_method_id: int, portfolio_id: int, sync_date: date, **kwargs: Any):
    portfolio = Portfolio.objects.get(id=portfolio_id)
    synchronization_method = PortfolioSynchronization.objects.get(id=synchronization_method_id)
    synchronization_method.synchronize(portfolio, sync_date, **kwargs)



from typing import Optional

from django.utils.translation import gettext_lazy as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class PortfolioDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Name"),
                dp.Field(key="currency", label="Currency"),
                dp.Field(key="portfolio_synchronization", label="Synchronization Method"),
                dp.Field(key="last_synchronization", label="Last Synchronization"),
                dp.Field(key="depends_on", label="Depends on"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["name", "currency"],
                ["portfolio_synchronization", "portfolio_synchronization"],
                [repeat_field(2, "dependent_portfolios")],
                [repeat_field(2, "trade_proposals_section")],
                [repeat_field(2, "instruments_section")],
                [repeat_field(2, "dependencyportfolios_section")],
                [repeat_field(2, "preferredclassification_section")],
            ],
            [
                create_simple_section(
                    "trade_proposals_section",
                    _("Trade Proposals"),
                    [["trade_proposals"]],
                    "trade_proposals",
                    collapsed=True,
                ),
                create_simple_section(
                    "instruments_section", _("Linked Instruments"), [["instruments"]], "instruments", collapsed=True
                ),
                create_simple_section(
                    "dependencyportfolios_section",
                    _("Dependency Portfolios"),
                    [["dependencyportfolios"]],
                    "dependencyportfolios",
                    collapsed=True,
                ),
                create_simple_section(
                    "preferredclassification_section",
                    _("Preferred Classification"),
                    [["preferredclassification"]],
                    "preferredclassification",
                    collapsed=True,
                ),
            ],
        )


class ModelPortfolioDisplayConfig(PortfolioDisplayConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Name"),
                dp.Field(key="currency", label="Currency"),
                dp.Field(key="portfolio_synchronization", label="Synchronization Method"),
                dp.Field(key="last_synchronization", label="Last Synchronization"),
                dp.Field(key="depends_on", label="Depends on"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["name", "currency"],
                [repeat_field(2, "portfolio_synchronization")],
                [repeat_field(2, "dependent_portfolios")],
                [repeat_field(2, "trade_proposals_section")],
                [repeat_field(2, "instruments_section")],
                [repeat_field(2, "dependencyportfolios_section")],
                [repeat_field(2, "preferredclassification_section")],
            ],
            [
                create_simple_section(
                    "trade_proposals_section",
                    _("Trade Proposals"),
                    [["trade_proposals"]],
                    "trade_proposals",
                    collapsed=True,
                ),
                create_simple_section(
                    "instruments_section", _("Linked Instruments"), [["instruments"]], "instruments", collapsed=True
                ),
                create_simple_section(
                    "dependencyportfolios_section",
                    _("Dependency Portfolios"),
                    [["dependencyportfolios"]],
                    "dependencyportfolios",
                    collapsed=True,
                ),
                create_simple_section(
                    "preferredclassification_section",
                    _("Preferred Classification"),
                    [["preferredclassification"]],
                    "preferredclassification",
                    collapsed=True,
                ),
            ],
        )


class PortfolioPortfolioThroughModelDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="dependency_portfolio", label="Dependency Portfolio"),
                dp.Field(key="type", label="Type"),
            ]
        )



from .adjustments import AdjustmentFactory
from .assets import AssetPositionFactory
from .claim import (
    ApprovedClaimFactory,
    ClaimFactory,
    NegativeClaimFactory,
    PositiveClaimFactory,
)
from .custodians import CustodianFactory
from .dividends import DividendTransactionsFactory
from .fees import FeesFactory
from wbfdm.factories.instrument_prices import InstrumentPriceFactory
from .portfolios import (
    InstrumentPortfolioThroughModelFactory,
    ModelPortfolioFactory,
    ModelPortfolioWithBaseProductFactory,
    PortfolioFactory,
)
from .portfolio_swing_pricings import PortfolioSwingPricingFactory
from .portfolio_cash_targets import PortfolioCashTargetFactory
from .portfolio_cash_flow import DailyPortfolioCashFlowFactory
from .product_groups import ProductGroupFactory, ProductGroupRepresentantFactory
from .products import IndexProductFactory, ProductFactory, WhiteLabelProductFactory
from .reconciliations import AccountReconciliationFactory, AccountReconciliationLineFactory
from .roles import ManagerPortfolioRoleFactory, ProductPortfolioRoleFactory
from .synchronization import (
    CrontabScheduleFactory,
    PeriodicTaskFactory,
    PortfolioSynchronizationFactory,
    PriceComputationFactory,
    SynchronizationTaskFactory,
)
from .trades import CustomerTradeFactory, TradeFactory, TradeProposalFactory
from .transactions import TransactionFactory
from .indexes import IndexFactory



import importlib
from collections.abc import Callable
from datetime import date, datetime, time
from json import loads
from typing import Any, Iterator, Optional

from celery import chord, group, shared_task
from celery.canvas import Signature, signature
from croniter import croniter, croniter_range
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone
from django_celery_beat.models import CrontabSchedule, PeriodicTask, cronexp


class SynchronizationTask(PeriodicTask):
    RELATIVE_TASK_MODULE_PATH = "wbportfolio.models.synchronization.synchronization.task"

    dependent_task = models.ForeignKey(
        "self", related_name="dependency_tasks", null=True, blank=True, on_delete=models.SET_NULL
    )
    import_path = models.CharField(max_length=512, verbose_name="Import Path", default="", blank=True)

    @property
    def cast_args(self) -> list[Any]:
        """
        transform the string args representation into list. We except this to become unnecessary when django beat move to jsonfield
        :return: list
        """
        return loads(self.args or "[]")

    @property
    def cast_kwargs(self) -> dict[Any, Any]:
        """
        transform the string kwargs representation into dictionary. We except this to become unnecessary when django beat move to jsonfield
        :return: list
        """
        return loads(self.kwargs or "{}")

    @property
    def _import_method(self) -> Callable[[Any], Any]:
        """
        If a custom task is specified, return the loaded module as callabck. Otherwise, returns the default synchronization
        function defined in `_default_callback
        :return: Callable
        """
        if import_path := self.import_path:
            synchronize_module = importlib.import_module(import_path)
            return synchronize_module.callback
        else:
            return self._default_callback

    def schedule_str(self, filter_daily: Optional[bool] = False) -> str:
        """
        Returns the crontab string representation. If fitler_daily is true, we cast the crontab so that the lowest frequency
        becomes daily.
        :param filter_daily: bool (optional)
        :return: crontab string representation
        """
        crontab_minute = cronexp(self.crontab.minute)
        crontab_hour = cronexp(self.crontab.hour)
        if filter_daily:
            if crontab_minute == "*":
                crontab_minute = "0"
            if crontab_hour == "*":
                crontab_hour = "0"
        return "{0} {1} {2} {3} {4}".format(
            crontab_minute,
            crontab_hour,
            cronexp(self.crontab.day_of_month),
            cronexp(self.crontab.month_of_year),
            cronexp(self.crontab.day_of_week),
        )

    def dates_range(self, from_date: date, to_date: date, filter_daily: Optional[bool] = False) -> Iterator[datetime]:
        """
        returns a list of valid dates given an interval and a specific crontab schedule.
        :param from_date: date
        :param to_date: date
        :param filter_daily: bool (optional)
        :return: list[date]
        """
        min_datetime = datetime.combine(from_date, time.min)
        max_datetime = datetime.combine(to_date, time.max)
        return croniter_range(min_datetime, max_datetime, self.schedule_str(filter_daily=filter_daily))

    def is_valid_date(self, sync_datetime: datetime) -> bool:
        """
        check wether a date is valid given the stored crontab schedule
        :param sync_datetime: datetime
        :return: bool
        """
        return croniter.match(self.schedule_str(), sync_datetime)

    def save(self, *args: Any, **kwargs: Any):
        self.task = self.RELATIVE_TASK_MODULE_PATH
        if not self.schedule:
            self.crontab, _ = CrontabSchedule.objects.get_or_create(
                minute="0",
                hour="*",
                day_of_week="*",
                day_of_month="*",
                month_of_year="*",
            )
        if not self.crontab:
            raise ValueError("Synchronization task supports only Crontab Schedule")
        if self.crontab.minute == "*":
            raise ValueError("The minimum crontab frequency supported is hourly.")
        super().save(*args, **kwargs)

    def _tasks_signature(self, *args, **kwargs: Any) -> Signature:
        """
        Gather all tasks that needs to run under this synchronization job as a list of celery signatures.
        This method is expected to be implemented at each inheriting class.
        :param args: list
        :param kwargs: dict
        :return: list[signature]
        """
        raise NotImplementedError()

    def _default_callback(self, *args: Any, **kwargs: Any) -> Any:
        """
        The default synchronization function executed if no custom task is provided for this synchronization object.
        This method is expected to be implemented at each inheriting class.
        :param args: list
        :param kwargs: dict
        :return: callable
        """
        raise NotImplementedError()

    def _get_kwargs(self) -> Any:
        """
        return the base keyword argument to be injected into the `_tasks_signature` method. Define as a standalone function
        in order to allow subclass definition.
        :return: dict
        """
        return {"task_execution_datetime": timezone.now()}

    def _end_task_signature(self, **kwargs: Any) -> Signature:
        """
        A synchronization object can defined a dependant task that will be executed at this end of all returned task from
        _tasks_signatures.
        This function returns the signature of this chained task.
        :param kwargs:
        :return: signature
        """
        if self.dependent_task:
            kwargs = {"override_execution_datetime_validity": True, **self.dependent_task.cast_kwargs, **kwargs}
            return signature(
                self.dependent_task.task, args=self.dependent_task.cast_args, kwargs=kwargs, immutable=True
            )
        return None

    def chord(self, **kwargs: Any) -> chord:
        """
        This function is the main entry point of the synchronization worklow. It is called from within the shared_task `task`
        and create the celery chord containing the list of tasks chained to the end task (if any)
        :param kwargs:
        :return: chord
        """
        kwargs = {**kwargs, **self._get_kwargs()}
        tasks = list(self._tasks_signature(**kwargs))
        if end_task := self._end_task_signature(**kwargs):
            return chord(tasks, end_task)
        return group(tasks)


@receiver(post_save, sender="wbportfolio.SynchronizationTask")
@receiver(post_save, sender="wbportfolio.PortfolioSynchronization")
@receiver(post_save, sender="wbportfolio.PriceComputation")
def post_save_synchronization_task(sender, instance: models.Model, created: bool, raw: bool, **kwargs: Any):
    """
    Ensure args attribute contains the necessary arguments to retrieve the calling job from within asynchronous task
    """
    if (created and not raw) or not instance.args:
        content_type = ContentType.objects.get_for_model(instance)
        instance.args = f'["{instance.id}", "{content_type.app_label}", "{content_type.model}"]'
        instance.save()


@shared_task
def task(synchronization_object_id: int, app_label: str, model: str, **kwargs: Any):
    synchronization_object = ContentType.objects.get(app_label=app_label, model=model).get_object_for_this_type(
        id=synchronization_object_id
    )
    synchronization_object.chord(**kwargs).apply_async()



from django.db import models
from wbfdm.models.instruments import InstrumentType

from .mixins.instruments import PMSInstrument


class Index(PMSInstrument):
    price_computation = models.ForeignKey(
        "wbportfolio.PriceComputation",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="indexes",
        verbose_name="Price Computation Method",
    )
    risk_scale = models.PositiveIntegerField(null=True, blank=True, verbose_name="Risk Scale")

    def pre_save(self):
        super().pre_save()
        self.instrument_type = InstrumentType.INDEX
        if "market_data" not in self.dl_parameters:
            # we default to the internal dataloader
            self.dl_parameters["market_data"] = {
                "path": "wbfdm.contrib.internal.dataloaders.market_data.MarketDataDataloader"
            }
        self.is_managed = True

    class Meta:
        verbose_name = "Index"
        verbose_name_plural = "Indexes"

    def compute_str(self):
        return f"{self.name}  ({self.ticker} - {self.isin})"



from contextlib import suppress
from datetime import date, timedelta

from celery import chain, chord
from django.db.models import ProtectedError, Q
from pandas.tseries.offsets import BDay
from tqdm import tqdm
from wbfdm.models import InstrumentPrice
from wbportfolio.models import Portfolio, Trade
from wbportfolio.models.portfolio import propagate_or_update_portfolio_assets_as_task
from wbportfolio.models.products import Product, update_outstanding_shares_as_task

from .fdm.tasks import *  # noqa


@shared_task()
def dummy_task():
    return True


@shared_task(queue="portfolio")
def periodically_update_outstanding_shares_for_active_products():
    qs = Product.active_objects.all()
    for product in tqdm(qs, total=qs.count()):
        update_outstanding_shares_as_task(product.id)


@shared_task(queue="portfolio")
def periodically_clean_marked_for_deletion_trades(max_allowed_iterations: int = 5):
    # Get all trade marked for deletion or pending and older than 7 days (i.e. After 7 days, we consider the pending trade obselete)
    qs = Trade.objects.filter(
        Q(marked_for_deletion=True) | (Q(pending=True) & Q(transaction_date__lt=date.today() - timedelta(days=7)))
    )
    i = 0

    # We try several times in case the trades deletion mechanism shifts the marked for deletion tag forwards
    while i < max_allowed_iterations and qs.exists():
        for t in qs:
            with suppress(ProtectedError):
                t.delete()
        qs = Trade.objects.filter(marked_for_deletion=True)
        i += 1


# Daily synchronization tasks.
# This tasks needs to be ran at maximum once a day in order to guarantee data consitency in
# case of change in change (e.g. reimport).
@shared_task(queue="portfolio")
def daily_instrument_price_statistics_synchronization(today: date = None, day_periods: int = 7):
    if not today:
        today = date.today()

    # We query for the last 7 days unsynch instrument prices.
    prices = (
        InstrumentPrice.objects.filter(
            date__gte=today - timedelta(days=day_periods), instrument__related_instruments__isnull=False
        )
        .filter(Q(sharpe_ratio__isnull=True) | Q(correlation__isnull=True) | Q(beta__isnull=True))
        .distinct()
    )
    objs = []
    for p in prices.iterator():
        p.compute_and_update_statistics()
        objs.append(p)
    InstrumentPrice.objects.bulk_update(objs, fields=["sharpe_ratio", "correlation", "beta"])


# A Task to run every day to update automatically the preferred classification
# per instrument of each wbportfolio containing assets.
@shared_task(queue="portfolio")
def update_preferred_classification_per_instrument_and_portfolio_as_task():
    for portfolio in Portfolio.tracked_objects.all():
        portfolio.update_preferred_classification_per_instrument()


# This task needs to run at fix interval. It will trigger the basic wbportfolio synchronization update:
# - Fetch for stainly price at t-1
# - propagate (or update) t-2 asset positions into t-1
# - Synchronize wbportfolio at t-1
# - Compute Instrument Price estimate at t-1


@shared_task(queue="portfolio")
def general_portfolio_synchronization_and_update(task_date=None):
    if not task_date:
        task_date = date.today()
    t_1 = (task_date - BDay(1)).date()
    t_2 = (t_1 - BDay(1)).date()

    # We propagate or update assets position from t-2 to t-1 (Needs stainly t-1 prices)
    subroutines_propagate_or_update_portfolio_assets = list()
    for portfolio in Portfolio.tracked_objects.all():
        if (
            portfolio.assets.filter(date=t_2).exists()
            and not portfolio.assets.filter(date=t_1, is_estimated=False).exists()
        ):
            subroutines_propagate_or_update_portfolio_assets.append(
                propagate_or_update_portfolio_assets_as_task.si(portfolio.id, t_2, t_1)
            )

    # # Synchronize wbportfolio at t-1 (needs propagated data)
    # subroutines_synchronize_computed_positions = list()
    # for method in PortfolioSynchronization.objects.all():
    #     for portfolio in method.portfolios.all():
    #         subroutines_synchronize_computed_positions.append(
    #             synchronize_portfolio_as_task.si(portfolio.id, t_1, synchronization_method_id=method.id)
    #         )
    #
    # # Compute price estimate at t-1 (needs synchronized wbportfolio)
    # subroutines_compute_price = list()
    # for method in PriceComputation.objects.all():
    #     for instrument in method.instruments.all():
    #         subroutines_compute_price.append(
    #             compute_price_as_task.si(instrument.id, t_1, price_computation_method_id=method.id)
    #         )

    subroutines_propagate_or_update_portfolio_assets_group = chord(
        subroutines_propagate_or_update_portfolio_assets, dummy_task.si()
    )

    chain(
        subroutines_propagate_or_update_portfolio_assets_group,
        # subroutines_synchronize_computed_positions_group,
        # subroutines_compute_price_group
    ).apply_async()



from django.contrib import admin
from wbportfolio.models import PortfolioSynchronization

from .admin import SynchronizationAdmin


@admin.register(PortfolioSynchronization)
class PortfolioSynchronizationModelAdmin(SynchronizationAdmin):
    fieldsets = (
        *SynchronizationAdmin.fieldsets,
        (
            "Synchronization",
            {
                "fields": ("import_path", "dependent_task", "is_automatic_validation"),
                "classes": ("extrapretty", "wide"),
            },
        ),
    )
